# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from pathlib import PureWindowsPath

from twisted.internet import error
from twisted.python.reflect import namedModule
from twisted.trial import unittest

from buildbot import config
from buildbot.process import remotetransfer
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import mercurial
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps


class TestMercurial(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def patch_workerVersionIsOlderThan(self, result):
        self.patch(mercurial.Mercurial, 'workerVersionIsOlderThan', lambda x, y, z: result)

    def test_no_repourl(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(mode="full")

    def test_incorrect_mode(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org', mode='invalid')

    def test_incorrect_method(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org', method='invalid')

    def test_incorrect_branchType(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org', branchType='invalid')

    def test_mode_full_clean(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_win32path(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            )
        )
        self.build.path_module = namedModule('ntpath')
        self.build.path_cls = PureWindowsPath
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file=r'wkdir\.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file=r'wkdir\.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_timeout(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                timeout=1,
                mode='full',
                method='clean',
                branchType='inrepo',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', timeout=1, command=['hg', '--verbose', '--version']).exit(
                0
            ),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', timeout=1, command=['hg', '--verbose', 'identify', '--branch']
            )
            .stdout('default')
            .exit(0),
            ExpectShell(
                workdir='wkdir', timeout=1, command=['hg', '--verbose', 'locate', 'set:added()']
            ).exit(1),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                timeout=1,
                command=['hg', '--verbose', 'parents', '--template', '{node}\\n'],
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                initial_stdin='patch',
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                initial_stdin='patch',
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch_fail(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                initial_stdin='patch',
            ).exit(1),
        )
        self.expect_outcome(result=FAILURE, state_string="update (failure)")
        return self.run_step()

    def test_mode_full_clean_no_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
                log_environ=True,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clobber(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clobber', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='fresh', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge', '--all'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh_no_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='fresh', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
                log_environ=True,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh_retry(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                mode='full',
                method='fresh',
                branchType='inrepo',
                retry=(0, 2),
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
                log_environ=True,
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_no_existing_repo_dirname(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='dirname'
            ),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),  # does not exist
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'update', '--clean']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_retry(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                mode='incremental',
                branchType='dirname',
                retry=(0, 1),
            ),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),  # does not exist
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'update', '--clean']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change_dirname(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org/',
                mode='incremental',
                branchType='dirname',
                defaultBranch='devel',
            ),
            {"branch": 'stable'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org/stable']
            ).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'hg',
                    '--verbose',
                    'clone',
                    '--noupdate',
                    'http://hg.mozilla.org/stable',
                    '.',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'update', '--clean']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_no_existing_repo_inrepo(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(1),  # does not exist
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),  # directory exists
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo_added_files(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),  # directory exists
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()'])
            .stdout('foo\nbar/baz\n')
            .exit(1),
            ExpectRmdir(dir=['wkdir/foo', 'wkdir/bar/baz'], log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo_added_files_old_rmdir(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            )
        )
        self.patch_workerVersionIsOlderThan(True)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),  # directory exists
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()'])
            .stdout('foo\nbar/baz\n')
            .exit(1),
            ExpectRmdir(dir='wkdir/foo', log_environ=True).exit(0),
            ExpectRmdir(dir='wkdir/bar/baz', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_given_revision(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            ),
            {"revision": 'abcdef01'},
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'abcdef01'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'abcdef01'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='incremental', branchType='inrepo'
            ),
            {"branch": 'stable'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'stable'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'clone', '--noupdate', 'http://hg.mozilla.org', '.'],
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'update', '--clean', '--rev', 'stable']
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change_no_clobberOnBranchChange(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                mode='incremental',
                branchType='inrepo',
                clobberOnBranchChange=False,
            ),
            {"branch": 'stable'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'stable'],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', 'locate', 'set:added()']).exit(
                1
            ),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'update', '--clean', '--rev', 'stable']
            ).exit(0),
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', 'parents', '--template', '{node}\\n']
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_env(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                mode='full',
                method='clean',
                branchType='inrepo',
                env={'abc': '123'},
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', '--version'], env={'abc': '123'}
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
                env={'abc': '123'},
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
                env={'abc': '123'},
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'identify', '--branch'],
                env={'abc': '123'},
            )
            .stdout('default')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'locate', 'set:added()'],
                env={'abc': '123'},
            ).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
                env={'abc': '123'},
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'parents', '--template', '{node}\\n'],
                env={'abc': '123'},
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_log_environ(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org',
                mode='full',
                method='clean',
                branchType='inrepo',
                logEnviron=False,
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['hg', '--verbose', '--version'], log_environ=False
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=False).exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=False).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', '--config', 'extensions.purge=', 'purge'],
                log_environ=False,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'pull', 'http://hg.mozilla.org', '--rev', 'default'],
                log_environ=False,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'identify', '--branch'],
                log_environ=False,
            )
            .stdout('default')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'locate', 'set:added()'],
                log_environ=False,
            ).exit(1),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'update', '--clean', '--rev', 'default'],
                log_environ=False,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['hg', '--verbose', 'parents', '--template', '{node}\\n'],
                log_environ=False,
            )
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_command_fails(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='fresh', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).exit(1)
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            mercurial.Mercurial(
                repourl='http://hg.mozilla.org', mode='full', method='clean', branchType='inrepo'
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['hg', '--verbose', '--version']).error(
                error.ConnectionLost()
            )
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import shutil

from twisted.internet import defer

from buildbot.process.buildstep import BuildStep
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps.transfer import DirectoryUpload
from buildbot.steps.transfer import FileDownload
from buildbot.steps.transfer import FileUpload
from buildbot.steps.transfer import MultipleFileUpload
from buildbot.steps.transfer import StringDownload
from buildbot.steps.worker import CompositeStepMixin
from buildbot.test.util.decorators import flaky
from buildbot.test.util.integration import RunMasterBase

# This integration test creates a master and worker environment
# and make sure the transfer steps are working

# When new protocols are added, make sure you update this test to exercise
# your proto implementation


class TransferStepsMasterPb(RunMasterBase):
    proto = "pb"

    @defer.inlineCallbacks
    def setup_config(self, bigfilename):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()
        # do a bunch of transfer to exercise the protocol
        f.addStep(StringDownload("filecontent", workerdest="dir/file1.txt"))
        f.addStep(StringDownload("filecontent2", workerdest="dir/file2.txt"))
        # create 8 MB file
        with open(bigfilename, 'w', encoding='utf-8') as o:
            buf = "xxxxxxxx" * 1024
            for _ in range(1000):
                o.write(buf)
        f.addStep(FileDownload(mastersrc=bigfilename, workerdest="bigfile.txt"))
        f.addStep(FileUpload(workersrc="dir/file2.txt", masterdest="master.txt"))
        f.addStep(FileDownload(mastersrc="master.txt", workerdest="dir/file3.txt"))
        f.addStep(DirectoryUpload(workersrc="dir", masterdest="dir"))
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def setup_config_glob(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.process.factory import BuildFactory

        class CustomStep(BuildStep, CompositeStepMixin):
            @defer.inlineCallbacks
            def run(self):
                content = yield self.getFileContentFromWorker(
                    "dir/file1.txt", abandonOnFailure=True
                )
                assert content == "filecontent"
                return SUCCESS

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(StringDownload("filecontent", workerdest="dir/file1.txt"))
        f.addStep(StringDownload("filecontent2", workerdest="dir/notafile1.txt"))
        f.addStep(StringDownload("filecontent2", workerdest="dir/only1.txt"))
        f.addStep(
            MultipleFileUpload(
                workersrcs=["dir/file*.txt", "dir/not*.txt", "dir/only?.txt"],
                masterdest="dest/",
                glob=True,
            )
        )
        f.addStep(CustomStep())
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def setup_config_single_step(self, step):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()

        f.addStep(step)
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    def readMasterDirContents(self, top):
        contents = {}
        for root, _, files in os.walk(top):
            for name in files:
                fn = os.path.join(root, name)
                with open(fn, encoding='utf-8') as f:
                    contents[fn] = f.read()
        return contents

    @flaky(bugNumber=4407, onPlatform='win32')
    @defer.inlineCallbacks
    def test_transfer(self):
        yield self.setup_config(bigfilename=self.mktemp())

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['results'], SUCCESS)
        dirContents = self.readMasterDirContents("dir")
        self.assertEqual(
            dirContents,
            {
                os.path.join('dir', 'file1.txt'): 'filecontent',
                os.path.join('dir', 'file2.txt'): 'filecontent2',
                os.path.join('dir', 'file3.txt'): 'filecontent2',
            },
        )

        # cleanup our mess (worker is cleaned up by parent class)
        shutil.rmtree("dir")
        os.unlink("master.txt")

    @defer.inlineCallbacks
    def test_globTransfer(self):
        yield self.setup_config_glob()
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['results'], SUCCESS)
        dirContents = self.readMasterDirContents("dest")
        self.assertEqual(
            dirContents,
            {
                os.path.join('dest', 'file1.txt'): 'filecontent',
                os.path.join('dest', 'notafile1.txt'): 'filecontent2',
                os.path.join('dest', 'only1.txt'): 'filecontent2',
            },
        )

        # cleanup
        shutil.rmtree("dest")

    @defer.inlineCallbacks
    def test_no_exist_file_upload(self):
        step = FileUpload(workersrc="dir/noexist_path", masterdest="master_dest")
        yield self.setup_config_single_step(step)

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['results'], FAILURE)
        res = yield self.checkBuildStepLogExist(build, "Cannot open file")
        self.assertTrue(res)
        self.assertFalse(os.path.exists("master_dest"))

    @defer.inlineCallbacks
    def test_no_exist_directory_upload(self):
        step = DirectoryUpload(workersrc="dir/noexist_path", masterdest="master_dest")
        yield self.setup_config_single_step(step)

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['results'], FAILURE)
        res = yield self.checkBuildStepLogExist(build, "Cannot read directory")
        self.assertTrue(res)
        self.assertFalse(os.path.exists("master_dest"))

    @defer.inlineCallbacks
    def test_no_exist_multiple_file_upload(self):
        step = MultipleFileUpload(workersrcs=["dir/noexist_path"], masterdest="master_dest")
        yield self.setup_config_single_step(step)

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['results'], FAILURE)
        res = yield self.checkBuildStepLogExist(build, "not available at worker")
        self.assertTrue(res)
        self.assertEqual(self.readMasterDirContents("master_dest"), {})


class TransferStepsMasterNull(TransferStepsMasterPb):
    proto = "null"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.trial import unittest

from buildbot.config.errors import ConfigErrors
from buildbot.config.errors import capture_config_errors
from buildbot.config.errors import error


class ConfigErrorsTest(unittest.TestCase):
    def test_constr(self):
        ex = ConfigErrors(['a', 'b'])
        self.assertEqual(ex.errors, ['a', 'b'])

    def test_addError(self):
        ex = ConfigErrors(['a'])
        ex.addError('c')
        self.assertEqual(ex.errors, ['a', 'c'])

    def test_nonempty(self):
        empty = ConfigErrors()
        full = ConfigErrors(['a'])
        self.assertTrue(not empty)
        self.assertFalse(not full)

    def test_error_raises(self):
        with self.assertRaises(ConfigErrors) as e:
            error("message")
        self.assertEqual(e.exception.errors, ["message"])

    def test_error_no_raise(self):
        with capture_config_errors() as errors:
            error("message")
        self.assertEqual(errors.errors, ["message"])

    def test_str(self):
        ex = ConfigErrors()
        self.assertEqual(str(ex), "")

        ex = ConfigErrors(["a"])
        self.assertEqual(str(ex), "a")

        ex = ConfigErrors(["a", "b"])
        self.assertEqual(str(ex), "a\nb")

        ex = ConfigErrors(["a"])
        ex.addError('c')
        self.assertEqual(str(ex), "a\nc")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Patch(Row):
    table = "patches"

    id_column = 'id'

    def __init__(
        self,
        id=None,
        patchlevel=0,
        patch_base64='aGVsbG8sIHdvcmxk',  # 'hello, world',
        patch_author=None,
        patch_comment=None,
        subdir=None,
    ):
        super().__init__(
            id=id,
            patchlevel=patchlevel,
            patch_base64=patch_base64,
            patch_author=patch_author,
            patch_comment=patch_comment,
            subdir=subdir,
        )


class NotSet:
    pass


class SourceStamp(Row):
    table = "sourcestamps"

    id_column = 'id'
    hashedColumns = [
        (
            'ss_hash',
            (
                'branch',
                'revision',
                'repository',
                'project',
                'codebase',
                'patchid',
            ),
        )
    ]

    def __init__(
        self,
        id=None,
        branch='master',
        revision=NotSet,
        patchid=None,
        repository='repo',
        codebase='',
        project='proj',
        created_at=89834834,
        ss_hash=None,
    ):
        if revision is NotSet:
            revision = f'rev-{id}'
        super().__init__(
            id=id,
            branch=branch,
            revision=revision,
            patchid=patchid,
            repository=repository,
            codebase=codebase,
            project=project,
            created_at=created_at,
            ss_hash=ss_hash,
        )



#! /usr/bin/python


import os.path
import time

import MySQLdb  # @UnresolvedImport
from twisted.cred import credentials
from twisted.internet import reactor
from twisted.python import log
from twisted.spread import pb

"""Based on the fakechanges.py contrib script"""


class ViewCvsPoller:
    def __init__(self):
        def _load_rc():
            import user

            ret = {}
            for line in open(os.path.join(user.home, ".cvsblamerc")).readlines():
                if line.find("=") != -1:
                    key, val = line.split("=")
                    ret[key.strip()] = val.strip()
            return ret

        # maybe add your own keys here db=xxx, user=xxx, passwd=xxx
        self.cvsdb = MySQLdb.connect("cvs", **_load_rc())
        # self.last_checkin = "2005-05-11" # for testing
        self.last_checkin = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())

    def get_changes(self):
        changes = []

        def empty_change():
            return {'who': None, 'files': [], 'comments': None}

        change = empty_change()

        cursor = self.cvsdb.cursor()
        cursor.execute(
            f"""SELECT whoid, descid, fileid, dirid, branchid, \
ci_when FROM checkins WHERE ci_when>='{self.last_checkin}'"""
        )
        last_checkin = None
        for whoid, descid, fileid, dirid, branchid, ci_when in cursor.fetchall():
            if branchid != 1:  # only head
                continue
            cursor.execute(f"""SELECT who from people where id={whoid}""")
            who = cursor.fetchone()[0]
            cursor.execute(f"""SELECT description from descs where id={descid}""")
            desc = cursor.fetchone()[0]
            cursor.execute(f"""SELECT file from files where id={fileid}""")
            filename = cursor.fetchone()[0]
            cursor.execute(f"""SELECT dir from dirs where id={dirid}""")
            dirname = cursor.fetchone()[0]
            if who == change["who"] and desc == change["comments"]:
                change["files"].append(f"{dirname}/{filename}")
            elif change["who"]:
                changes.append(change)
                change = empty_change()
            else:
                change["who"] = who
                change["files"].append(f"{dirname}/{filename}")
                change["comments"] = desc
            if last_checkin is None or ci_when > last_checkin:
                last_checkin = ci_when
        if last_checkin:
            self.last_checkin = last_checkin
        return changes


poller = ViewCvsPoller()


def error(*args):
    log.err()
    reactor.stop()


def poll_changes(remote):
    print("GET CHANGES SINCE", poller.last_checkin, end=' ')
    changes = poller.get_changes()
    for change in changes:
        print(change["who"], "\n *", "\n * ".join(change["files"]))
        change['src'] = 'cvs'
        remote.callRemote('addChange', change).addErrback(error)
    print()
    reactor.callLater(60, poll_changes, remote)


factory = pb.PBClientFactory()
reactor.connectTCP("localhost", 9999, factory)
deferred = factory.login(credentials.UsernamePassword("change", "changepw"))
deferred.addCallback(poll_changes).addErrback(error)

reactor.run()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import re
from typing import NamedTuple

# The regex is matching more than it should and is not intended to be an url validator.
# It is intended to efficiently and reliably extract information from the various examples
# that are described in the unit tests.

_giturlmatcher = re.compile(
    r'(?P<proto>(https?://|ssh://|git://|))'
    r'((?P<user>[^:@]*)(:(?P<password>.*))?@)?'
    r'(?P<domain>[^\/:]+)(:((?P<port>[0-9]+)/)?|/)'
    r'((?P<owner>.+)/)?(?P<repo>[^/]+?)(\.git)?$'
)


class GitUrl(NamedTuple):
    proto: str
    user: str | None
    password: str | None
    domain: str
    port: int | None
    owner: str | None
    repo: str


def giturlparse(url: str) -> GitUrl | None:
    res = _giturlmatcher.match(url)
    if res is None:
        return None

    port = res.group("port")
    if port is not None:
        port = int(port)
    proto = res.group("proto")
    if proto:
        proto = proto[:-3]
    else:
        proto = 'ssh'  # implicit proto is ssh
    return GitUrl(
        proto=proto,
        user=res.group('user'),
        password=res.group('password'),
        domain=res.group("domain"),
        port=port,
        owner=res.group('owner'),
        repo=res.group('repo'),
    )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import sourcestamps
from buildbot.test import fakedb
from buildbot.test.util import endpoint


class SourceStampEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = sourcestamps.SourceStampEndpoint
    resourceTypeClass = sourcestamps.SourceStamp

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=13, branch='oak'),
            fakedb.Patch(
                id=99,
                patch_base64='aGVsbG8sIHdvcmxk',
                patch_author='bar',
                patch_comment='foo',
                subdir='/foo',
                patchlevel=3,
            ),
            fakedb.SourceStamp(id=14, patchid=99, branch='poplar'),
        ])

    @defer.inlineCallbacks
    def test_get_existing(self):
        sourcestamp = yield self.callGet(('sourcestamps', 13))

        self.validateData(sourcestamp)
        self.assertEqual(sourcestamp['branch'], 'oak')
        self.assertEqual(sourcestamp['patch'], None)

    @defer.inlineCallbacks
    def test_get_existing_patch(self):
        sourcestamp = yield self.callGet(('sourcestamps', 14))

        self.validateData(sourcestamp)
        self.assertEqual(sourcestamp['branch'], 'poplar')
        self.assertEqual(
            sourcestamp['patch'],
            {
                'patchid': 99,
                'author': 'bar',
                'body': b'hello, world',
                'comment': 'foo',
                'level': 3,
                'subdir': '/foo',
            },
        )

    @defer.inlineCallbacks
    def test_get_missing(self):
        sourcestamp = yield self.callGet(('sourcestamps', 99))

        self.assertEqual(sourcestamp, None)


class SourceStampsEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = sourcestamps.SourceStampsEndpoint
    resourceTypeClass = sourcestamps.SourceStamp

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=30, reason="foo", submitted_at=1300305712, results=-1),
            fakedb.SourceStamp(id=13),
            fakedb.SourceStamp(id=14),
            fakedb.SourceStamp(id=15),
            fakedb.BuildsetSourceStamp(sourcestampid=13, buildsetid=30),
            fakedb.BuildsetSourceStamp(sourcestampid=14, buildsetid=30),
        ])

    @defer.inlineCallbacks
    def test_get(self):
        sourcestamps = yield self.callGet(('sourcestamps',))

        for m in sourcestamps:
            self.validateData(m)

        self.assertEqual(sorted([m['ssid'] for m in sourcestamps]), [13, 14, 15])

    @defer.inlineCallbacks
    def test_get_by_buildsetid_no_buildset(self):
        sourcestamps = yield self.callGet(("buildsets", 101, "sourcestamps"))
        self.assertEqual(sourcestamps, [])

    @defer.inlineCallbacks
    def test_get_by_buildsetid(self):
        sourcestamps = yield self.callGet(("buildsets", 30, "sourcestamps"))

        for m in sourcestamps:
            self.validateData(m)

        self.assertEqual(sorted([m['ssid'] for m in sourcestamps]), [13, 14])


class SourceStamp(unittest.TestCase):
    pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import builtins
import os
import re
import textwrap
from importlib.util import find_spec
from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest
from zope.interface import implementer

from buildbot import config
from buildbot import configurators
from buildbot import interfaces
from buildbot import locks
from buildbot import revlinks
from buildbot import worker
from buildbot.changes import base as changes_base
from buildbot.config.errors import capture_config_errors
from buildbot.config.master import FileLoader
from buildbot.config.master import loadConfigDict
from buildbot.process import factory
from buildbot.process import properties
from buildbot.process.codebase import Codebase
from buildbot.process.project import Project
from buildbot.schedulers import base as schedulers_base
from buildbot.test.util import dirs
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.warnings import assertNotProducesWarnings
from buildbot.test.util.warnings import assertProducesWarning
from buildbot.util import service
from buildbot.warnings import ConfigWarning
from buildbot.warnings import DeprecatedApiWarning

HAS_ZSTD = find_spec('zstandard') is not None

global_defaults = {
    "title": 'Buildbot',
    "titleURL": 'http://buildbot.net/',
    "buildbotURL": 'http://localhost:8080/',
    "logCompressionLimit": 4096,
    "logCompressionMethod": 'zstd' if HAS_ZSTD else 'gz',
    "logEncoding": 'utf-8',
    "logMaxTailSize": None,
    "logMaxSize": None,
    "properties": properties.Properties(),
    "collapseRequests": None,
    "prioritizeBuilders": None,
    "select_next_worker": None,
    "protocols": {},
    "multiMaster": False,
    "manhole": None,
    # in unit tests we default to None, but normally defaults to 'basic'
    "buildbotNetUsageData": None,
    "www": {
        "port": None,
        "plugins": {},
        "auth": {'name': 'NoAuth'},
        "authz": {},
        "avatar_methods": {'name': 'gravatar'},
        "logfileName": 'http.log',
    },
}


class FakeChangeSource(changes_base.ChangeSource):
    def __init__(self):
        super().__init__(name='FakeChangeSource')


@implementer(interfaces.IScheduler)
class FakeScheduler:
    def __init__(self, name):
        self.name = name


class FakeBuilder:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


@implementer(interfaces.IWorker)
class FakeWorker:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    @property
    def workername(self):
        return self.name


@implementer(interfaces.IMachine)
class FakeMachine:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class ConfigLoaderTests(ConfigErrorsMixin, dirs.DirsMixin, unittest.SynchronousTestCase):
    def setUp(self):
        self.basedir = os.path.abspath('basedir')
        self.filename = os.path.join(self.basedir, 'test.cfg')
        self.patch(config.master, "get_is_in_unit_tests", lambda: False)

        return self.setUpDirs('basedir')

    def install_config_file(self, config_file, other_files=None):
        if other_files is None:
            other_files = {}
        config_file = textwrap.dedent(config_file)
        with open(os.path.join(self.basedir, self.filename), "w", encoding='utf-8') as f:
            f.write(config_file)
        for file, contents in other_files.items():
            with open(file, "w", encoding='utf-8') as f:
                f.write(contents)

    def test_loadConfig_missing_file(self):
        with self.assertRaisesConfigError(re.compile("configuration file .* does not exist")):
            loadConfigDict(self.basedir, self.filename)

    def test_loadConfig_missing_basedir(self):
        with self.assertRaisesConfigError(re.compile("basedir .* does not exist")):
            loadConfigDict(os.path.join(self.basedir, 'NO'), 'test.cfg')

    def test_loadConfig_open_error(self):
        """
        Check that loadConfig() raises correct ConfigError exception in cases
        when configure file is found, but we fail to open it.
        """

        def raise_IOError(*args, **kwargs):
            raise OSError("error_msg")

        self.install_config_file('#dummy')

        # override build-in open() function to always rise IOError
        self.patch(builtins, "open", raise_IOError)

        # check that we got the expected ConfigError exception
        with self.assertRaisesConfigError(
            re.compile("unable to open configuration file .*: error_msg")
        ):
            loadConfigDict(self.basedir, self.filename)

    def test_loadConfig_parse_error(self):
        self.install_config_file('def x:\nbar')
        with self.assertRaisesConfigError(
            re.compile("encountered a SyntaxError while parsing config file:")
        ):
            loadConfigDict(self.basedir, self.filename)

    def test_loadConfig_eval_ConfigError(self):
        self.install_config_file("""\
                from buildbot import config
                BuildmasterConfig = { 'multiMaster': True }
                config.error('oh noes!')""")
        with self.assertRaisesConfigError("oh noes"):
            loadConfigDict(self.basedir, self.filename)

    def test_loadConfig_eval_otherError(self):
        self.install_config_file("""\
                from buildbot import config
                BuildmasterConfig = { 'multiMaster': True }
                raise ValueError('oh noes')""")
        with self.assertRaisesConfigError(
            "error while parsing config file: oh noes (traceback in logfile)"
        ):
            loadConfigDict(self.basedir, self.filename)

        [error] = self.flushLoggedErrors(ValueError)
        self.assertEqual(error.value.args, ("oh noes",))

    def test_loadConfig_no_BuildmasterConfig(self):
        self.install_config_file('x=10')
        with self.assertRaisesConfigError("does not define 'BuildmasterConfig'"):
            loadConfigDict(self.basedir, self.filename)

    def test_loadConfig_with_local_import(self):
        self.install_config_file(
            """\
                from subsidiary_module import x
                BuildmasterConfig = dict(x=x)
                """,
            {'basedir/subsidiary_module.py': "x = 10"},
        )
        _, rv = loadConfigDict(self.basedir, self.filename)
        self.assertEqual(rv, {'x': 10})


class MasterConfigTests(ConfigErrorsMixin, dirs.DirsMixin, unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.basedir = os.path.abspath('basedir')
        self.filename = os.path.join(self.basedir, 'test.cfg')
        return self.setUpDirs('basedir')

    # utils

    def patch_load_helpers(self):
        # patch out all of the "helpers" for loadConfig with null functions
        for n in dir(config.master.MasterConfig):
            if n.startswith('load_'):
                typ = 'loader'
            elif n.startswith('check_'):
                typ = 'checker'
            else:
                continue

            v = getattr(config.master.MasterConfig, n)
            if callable(v):
                if typ == 'loader':
                    self.patch(
                        config.master.MasterConfig,
                        n,
                        mock.Mock(side_effect=lambda filename, config_dict: None),
                    )
                else:
                    self.patch(config.master.MasterConfig, n, mock.Mock(side_effect=lambda: None))

    def install_config_file(self, config_file, other_files=None):
        if other_files is None:
            other_files = {}
        config_file = textwrap.dedent(config_file)
        with open(os.path.join(self.basedir, self.filename), "w", encoding='utf-8') as f:
            f.write(config_file)
        for file, contents in other_files.items():
            with open(file, "w", encoding='utf-8') as f:
                f.write(contents)

    # tests

    def test_defaults(self):
        cfg = config.master.MasterConfig()
        expected = {
            # validation,
            "db": config.master.DBConfig(db_url='sqlite:///state.sqlite'),
            "mq": {"type": 'simple'},
            "metrics": None,
            "caches": {"Changes": 10, "Builds": 15},
            "schedulers": {},
            "builders": [],
            "workers": [],
            "change_sources": [],
            "status": [],
            "user_managers": [],
            "revlink": revlinks.default_revlink_matcher,
        }
        expected.update(global_defaults)
        expected['buildbotNetUsageData'] = 'basic'
        got = {attr: getattr(cfg, attr) for attr, exp in expected.items()}
        got = interfaces.IConfigured(got).getConfigDict()
        expected = interfaces.IConfigured(expected).getConfigDict()
        self.assertEqual(got, expected)

    def test_defaults_validation(self):
        # re's aren't comparable, but we can make sure the keys match
        cfg = config.master.MasterConfig()
        self.assertEqual(
            sorted(cfg.validation.keys()),
            sorted([
                'branch',
                'revision',
                'property_name',
                'property_value',
            ]),
        )

    def test_loadConfig_eval_ConfigErrors(self):
        # We test a config that has embedded errors, as well
        # as semantic errors that get added later. If an exception is raised
        # prematurely, then the semantic errors wouldn't get reported.
        self.install_config_file("""\
                from buildbot import config
                BuildmasterConfig = {}
                config.error('oh noes!')
                config.error('noes too!')""")

        with capture_config_errors() as errors:
            FileLoader(self.basedir, self.filename).loadConfig()

        self.assertEqual(
            errors.errors,
            ['oh noes!', 'noes too!', 'no workers are configured', 'no builders are configured'],
        )

    def test_loadConfig_unknown_key(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict(foo=10)
                """)
        with self.assertRaisesConfigError("Unknown BuildmasterConfig key foo"):
            FileLoader(self.basedir, self.filename).loadConfig()

    def test_loadConfig_unknown_keys(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict(foo=10, bar=20)
                """)
        with self.assertRaisesConfigError("Unknown BuildmasterConfig keys bar, foo"):
            FileLoader(self.basedir, self.filename).loadConfig()

    def test_loadConfig_success(self):
        self.patch_load_helpers()
        self.install_config_file("""\
                BuildmasterConfig = dict()
                """)
        rv = FileLoader(self.basedir, self.filename).loadConfig()
        self.assertIsInstance(rv, config.master.MasterConfig)

        # make sure all of the loaders and checkers are called
        self.assertTrue(rv.load_global.called)
        self.assertTrue(rv.load_validation.called)
        self.assertTrue(rv.load_dbconfig.called)
        self.assertTrue(rv.load_metrics.called)
        self.assertTrue(rv.load_caches.called)
        self.assertTrue(rv.load_schedulers.called)
        self.assertTrue(rv.load_builders.called)
        self.assertTrue(rv.load_workers.called)
        self.assertTrue(rv.load_change_sources.called)
        self.assertTrue(rv.load_machines.called)
        self.assertTrue(rv.load_user_managers.called)

        self.assertTrue(rv.check_single_master.called)
        self.assertTrue(rv.check_schedulers.called)
        self.assertTrue(rv.check_locks.called)
        self.assertTrue(rv.check_builders.called)
        self.assertTrue(rv.check_ports.called)
        self.assertTrue(rv.check_machines.called)

    def test_preChangeGenerator(self):
        cfg = config.master.MasterConfig()
        self.assertEqual(
            {
                'author': None,
                'files': None,
                'comments': None,
                'revision': None,
                'when_timestamp': None,
                'branch': None,
                'category': None,
                'revlink': '',
                'properties': {},
                'repository': '',
                'project': '',
                'codebase': None,
            },
            cfg.preChangeGenerator(),
        )


class MasterConfig_loaders(ConfigErrorsMixin, unittest.TestCase):
    filename = 'test.cfg'

    def setUp(self):
        self.cfg = config.master.MasterConfig()

    # utils

    def assertResults(self, **expected):
        got = {attr: getattr(self.cfg, attr) for attr, exp in expected.items()}
        got = interfaces.IConfigured(got).getConfigDict()
        expected = interfaces.IConfigured(expected).getConfigDict()

        self.assertEqual(got, expected)

    # tests

    def test_load_global_defaults(self):
        self.maxDiff = None
        self.cfg.load_global(self.filename, {})
        self.assertResults(**global_defaults)

    def test_load_global_string_param_not_string(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {"title": 10})
        self.assertConfigError(errors, 'must be a string')

    def test_load_global_int_param_not_int(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'changeHorizon': 'yes'})

        self.assertConfigError(errors, 'must be an int')

    def test_load_global_protocols_not_dict(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'protocols': "test"})
        self.assertConfigError(errors, "c['protocols'] must be dict")

    def test_load_global_protocols_key_int(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'protocols': {321: {"port": 123}}})

        self.assertConfigError(errors, "c['protocols'] keys must be strings")

    def test_load_global_protocols_value_not_dict(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'protocols': {"pb": 123}})

        self.assertConfigError(errors, "c['protocols']['pb'] must be a dict")

    def do_test_load_global(self, config_dict, **expected):
        self.cfg.load_global(self.filename, config_dict)
        self.assertResults(**expected)

    def test_load_global_title(self):
        self.do_test_load_global({"title": 'hi'}, title='hi')

    def test_load_global_title_too_long(self):
        with assertProducesWarning(ConfigWarning, message_pattern=r"Title is too long"):
            self.do_test_load_global({"title": "Very very very very very long title"})

    def test_load_global_titleURL(self):
        self.do_test_load_global({"titleURL": 'hi'}, titleURL='hi/')

    def test_load_global_buildbotURL(self):
        self.do_test_load_global({"buildbotURL": 'hey'}, buildbotURL='hey/')

    def test_load_global_changeHorizon(self):
        self.do_test_load_global({"changeHorizon": 10}, changeHorizon=10)

    def test_load_global_changeHorizon_none(self):
        self.do_test_load_global({"changeHorizo": None}, changeHorizon=None)

    def test_load_global_buildbotNetUsageData(self):
        self.patch(config.master, "get_is_in_unit_tests", lambda: False)
        with assertProducesWarning(
            ConfigWarning,
            message_pattern=r"`buildbotNetUsageData` is not configured and defaults to basic.",
        ):
            self.do_test_load_global({})

    def test_load_global_logCompressionLimit(self):
        self.do_test_load_global({"logCompressionLimit": 10}, logCompressionLimit=10)

    def test_load_global_logCompressionMethod(self):
        self.do_test_load_global({"logCompressionMethod": 'bz2'}, logCompressionMethod='bz2')

    def test_load_global_logCompressionMethod_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'logCompressionMethod': 'foo'})

        self.assertConfigError(
            errors, "c['logCompressionMethod'] must be 'raw', 'bz2', 'gz', 'lz4', 'br' or 'zstd'"
        )

    def test_load_global_codebaseGenerator(self):
        func = lambda _: "dummy"
        self.do_test_load_global({"codebaseGenerator": func}, codebaseGenerator=func)

    def test_load_global_codebaseGenerator_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'codebaseGenerator': 'dummy'})

        self.assertConfigError(
            errors, "codebaseGenerator must be a callable accepting a dict and returning a str"
        )

    def test_load_global_logMaxSize(self):
        self.do_test_load_global({"logMaxSize": 123}, logMaxSize=123)

    def test_load_global_logMaxTailSize(self):
        self.do_test_load_global({"logMaxTailSize": 123}, logMaxTailSize=123)

    def test_load_global_logEncoding(self):
        self.do_test_load_global({"logEncoding": 'latin-2'}, logEncoding='latin-2')

    def test_load_global_properties(self):
        exp = properties.Properties()
        exp.setProperty('x', 10, self.filename)
        self.do_test_load_global({"properties": {"x": 10}}, properties=exp)

    def test_load_global_properties_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'properties': 'yes'})

        self.assertConfigError(errors, "must be a dictionary")

    def test_load_global_collapseRequests_bool(self):
        self.do_test_load_global({"collapseRequests": False}, collapseRequests=False)

    def test_load_global_collapseRequests_callable(self):
        callable = lambda: None
        self.do_test_load_global({"collapseRequests": callable}, collapseRequests=callable)

    def test_load_global_collapseRequests_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'collapseRequests': 'yes'})

        self.assertConfigError(errors, "must be a callable, True, or False")

    def test_load_global_prioritizeBuilders_callable(self):
        callable = lambda: None
        self.do_test_load_global({"prioritizeBuilders": callable}, prioritizeBuilders=callable)

    def test_load_global_prioritizeBuilders_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'prioritizeBuilders': 'yes'})

        self.assertConfigError(errors, "must be a callable")

    def test_load_global_select_next_worker_callable(self):
        callable = lambda: None
        self.do_test_load_global({"select_next_worker": callable}, select_next_worker=callable)

    def test_load_global_select_next_worker_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {"select_next_worker": "yes"})

        self.assertConfigError(errors, "must be a callable")

    def test_load_global_protocols_str(self):
        self.do_test_load_global(
            {"protocols": {'pb': {'port': 'udp:123'}}}, protocols={'pb': {'port': 'udp:123'}}
        )

    def test_load_global_multiMaster(self):
        self.do_test_load_global({"multiMaster": 1}, multiMaster=1)

    def test_load_global_manhole(self):
        mh = mock.Mock(name='manhole')
        self.do_test_load_global({"manhole": mh}, manhole=mh)

    def test_load_global_revlink_callable(self):
        callable = lambda: None
        self.do_test_load_global({"revlink": callable}, revlink=callable)

    def test_load_global_revlink_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_global(self.filename, {'revlink': ''})

        self.assertConfigError(errors, "must be a callable")

    def test_load_validation_defaults(self):
        self.cfg.load_validation(self.filename, {})
        self.assertEqual(
            sorted(self.cfg.validation.keys()),
            sorted([
                'branch',
                'revision',
                'property_name',
                'property_value',
            ]),
        )

    def test_load_validation_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_validation(self.filename, {'validation': 'plz'})

        self.assertConfigError(errors, "must be a dictionary")

    def test_load_validation_unk_keys(self):
        with capture_config_errors() as errors:
            self.cfg.load_validation(self.filename, {'validation': {'users': '.*'}})

        self.assertConfigError(errors, "unrecognized validation key(s)")

    def test_load_validation(self):
        r = re.compile('.*')
        self.cfg.load_validation(self.filename, {"validation": {"branch": r}})
        self.assertEqual(self.cfg.validation['branch'], r)
        # check that defaults are still around
        self.assertIn('revision', self.cfg.validation)

    def test_load_db_defaults(self):
        self.cfg.load_dbconfig(self.filename, {})
        self.assertResults(
            db=config.master.DBConfig(db_url='sqlite:///state.sqlite', engine_kwargs={})
        )

    def test_load_db_db_url(self):
        self.cfg.load_dbconfig(self.filename, {"db_url": 'abcd'})
        self.assertResults(db=config.master.DBConfig(db_url='abcd'))

    def test_load_db_dict(self):
        self.cfg.load_dbconfig(self.filename, {'db': {'db_url': 'abcd'}})
        self.assertResults(db=config.master.DBConfig(db_url='abcd'))

    def test_load_db_unk_keys(self):
        with capture_config_errors() as errors:
            self.cfg.load_dbconfig(self.filename, {'db': {'db_url': 'abcd', 'bar': 'bar'}})

        self.assertConfigError(errors, "unrecognized keys in")

    def test_load_mq_defaults(self):
        self.cfg.load_mq(self.filename, {})
        self.assertResults(mq={"type": 'simple'})

    def test_load_mq_explicit_type(self):
        self.cfg.load_mq(self.filename, {"mq": {"type": 'simple'}})
        self.assertResults(mq={"type": 'simple'})

    def test_load_mq_unk_type(self):
        with capture_config_errors() as errors:
            self.cfg.load_mq(self.filename, {'mq': {'type': 'foo'}})

        self.assertConfigError(errors, "mq type 'foo' is not known")

    def test_load_mq_unk_keys(self):
        with capture_config_errors() as errors:
            self.cfg.load_mq(self.filename, {'mq': {'bar': 'bar'}})

        self.assertConfigError(errors, "unrecognized keys in")

    def test_load_metrics_defaults(self):
        self.cfg.load_metrics(self.filename, {})
        self.assertResults(metrics=None)

    def test_load_metrics_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_metrics(self.filename, {'metrics': 13})

        self.assertConfigError(errors, "must be a dictionary")

    def test_load_metrics(self):
        self.cfg.load_metrics(self.filename, {"metrics": {"foo": 1}})
        self.assertResults(metrics={"foo": 1})

    def test_load_caches_defaults(self):
        self.cfg.load_caches(self.filename, {})
        self.assertResults(caches={"Changes": 10, "Builds": 15})

    def test_load_caches_invalid(self):
        with capture_config_errors() as errors:
            self.cfg.load_caches(self.filename, {'caches': 13})

        self.assertConfigError(errors, "must be a dictionary")

    def test_load_caches_buildCacheSize(self):
        self.cfg.load_caches(self.filename, {"buildCacheSize": 13})
        self.assertResults(caches={"Builds": 13, "Changes": 10})

    def test_load_caches_buildCacheSize_and_caches(self):
        with capture_config_errors() as errors:
            self.cfg.load_caches(self.filename, {'buildCacheSize': 13, 'caches': {'builds': 11}})

        self.assertConfigError(errors, "cannot specify")

    def test_load_caches_changeCacheSize(self):
        self.cfg.load_caches(self.filename, {"changeCacheSize": 13})
        self.assertResults(caches={"Changes": 13, "Builds": 15})

    def test_load_caches_changeCacheSize_and_caches(self):
        with capture_config_errors() as errors:
            self.cfg.load_caches(self.filename, {'changeCacheSize': 13, 'caches': {'changes': 11}})

        self.assertConfigError(errors, "cannot specify")

    def test_load_caches(self):
        self.cfg.load_caches(self.filename, {"caches": {"foo": 1}})
        self.assertResults(caches={"Changes": 10, "Builds": 15, "foo": 1})

    def test_load_caches_not_int_err(self):
        """
        Test that non-integer cache sizes are not allowed.
        """
        with capture_config_errors() as errors:
            self.cfg.load_caches(self.filename, {'caches': {'foo': "1"}})

        self.assertConfigError(errors, "value for cache size 'foo' must be an integer")

    def test_load_caches_to_small_err(self):
        """
        Test that cache sizes less then 1 are not allowed.
        """
        with capture_config_errors() as errors:
            self.cfg.load_caches(self.filename, {'caches': {'Changes': -12}})

        self.assertConfigError(errors, "'Changes' cache size must be at least 1, got '-12'")

    def test_load_schedulers_defaults(self):
        self.cfg.load_schedulers(self.filename, {})
        self.assertResults(schedulers={})

    def test_load_schedulers_not_list(self):
        with capture_config_errors() as errors:
            self.cfg.load_schedulers(self.filename, {'schedulers': {}})

        self.assertConfigError(errors, "must be a list of")

    def test_load_schedulers_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_schedulers(self.filename, {'schedulers': [mock.Mock()]})

        self.assertConfigError(errors, "must be a list of")

    def test_load_schedulers_dupe(self):
        with capture_config_errors() as errors:
            sch1 = FakeScheduler(name='sch')
            sch2 = FakeScheduler(name='sch')
            self.cfg.load_schedulers(self.filename, {'schedulers': [sch1, sch2]})
        self.assertConfigError(errors, "scheduler name 'sch' used multiple times")

    def test_load_schedulers(self):
        sch = schedulers_base.ReconfigurableBaseScheduler(name='sch', builderNames=["a"])
        self.cfg.load_schedulers(self.filename, {"schedulers": [sch]})
        self.assertResults(schedulers={"sch": sch})

    def test_load_builders_defaults(self):
        self.cfg.load_builders(self.filename, {})
        self.assertResults(builders=[])

    def test_load_builders_not_list(self):
        with capture_config_errors() as errors:
            self.cfg.load_builders(self.filename, {'builders': {}})

        self.assertConfigError(errors, "must be a list")

    def test_load_builders_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_builders(self.filename, {'builders': [mock.Mock()]})
        self.assertConfigError(errors, "is not a builder config (in c['builders']")

    def test_load_builders(self):
        bldr = config.BuilderConfig(name='x', factory=factory.BuildFactory(), workername='x')
        self.cfg.load_builders(self.filename, {"builders": [bldr]})
        self.assertResults(builders=[bldr])

    def test_load_builders_dict(self):
        bldr = {"name": 'x', "factory": factory.BuildFactory(), "workername": 'x'}
        self.cfg.load_builders(self.filename, {"builders": [bldr]})
        self.assertIsInstance(self.cfg.builders[0], config.BuilderConfig)
        self.assertEqual(self.cfg.builders[0].name, 'x')

    def test_load_builders_abs_builddir(self):
        bldr = {
            "name": 'x',
            "factory": factory.BuildFactory(),
            "workername": 'x',
            "builddir": os.path.abspath('.'),
        }
        self.cfg.load_builders(self.filename, {"builders": [bldr]})
        self.assertEqual(len(self.flushWarnings([self.cfg.load_builders])), 1)

    def test_load_codebases_defaults(self):
        self.cfg.load_codebases(self.filename, {})
        self.assertResults(codebases=[])

    def test_load_codebases_not_list(self):
        with capture_config_errors() as errors:
            self.cfg.load_codebases(self.filename, {'codebases': {}})
        self.assertConfigError(errors, "must be a list")

    def test_load_codebases_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_codebases(self.filename, {'codebases': [mock.Mock()]})

        self.assertConfigError(errors, "is not a codebase config")

    def test_load_codebases_not_project(self):
        with capture_config_errors() as errors:
            self.cfg.load_codebases(
                self.filename, {'codebases': [Codebase(name='codebase', project='project')]}
            )

        self.assertConfigError(errors, "includes unknown project")

    def test_load_codebases_valid(self):
        cfg = {
            'projects': [Project(name='project')],
            'codebases': [Codebase(name='codebase', project='project')],
        }

        self.cfg.load_projects(self.filename, cfg)
        self.cfg.load_codebases(self.filename, cfg)

        self.assertResults(codebases=[Codebase(name='codebase', project='project')])

    def test_load_workers_defaults(self):
        self.cfg.load_workers(self.filename, {})
        self.assertResults(workers=[])

    def test_load_workers_not_list(self):
        with capture_config_errors() as errors:
            self.cfg.load_workers(self.filename, {'workers': {}})
        self.assertConfigError(errors, "must be a list")

    def test_load_workers_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_workers(self.filename, {'workers': [mock.Mock()]})

        self.assertConfigError(errors, "must be a list of")

    @parameterized.expand(['debug', 'change', 'status'])
    def test_load_workers_reserved_names(self, worker_name):
        with capture_config_errors() as errors:
            self.cfg.load_workers(self.filename, {'workers': [worker.Worker(worker_name, 'x')]})

        self.assertConfigError(errors, "is reserved")

    @parameterized.expand([
        ('initial_digits', "123_text_text"),
        ("spaces", "text text"),
        ("slash", "a/b"),
        ("dot", "a.b"),
    ])
    def test_load_workers_not_identifiers(self, name, worker_name):
        with capture_config_errors() as errors:
            self.cfg.load_workers(self.filename, {'workers': [worker.Worker(worker_name, 'x')]})

        self.assertConfigError(errors, "is not an identifier")

    def test_load_workers_too_long(self):
        with capture_config_errors() as errors:
            name = "a" * 51
            self.cfg.load_workers(self.filename, {'workers': [worker.Worker(name, 'x')]})

        self.assertConfigError(errors, "is longer than")

    def test_load_workers_empty(self):
        with capture_config_errors() as errors:
            name = ""
            self.cfg.load_workers(self.filename, {'workers': [worker.Worker(name, 'x')]})
            errors.errors[:] = errors.errors[1:2]  # only get necessary error

        self.assertConfigError(errors, "cannot be an empty string")

    def test_load_workers(self):
        wrk = worker.Worker('foo', 'x')
        self.cfg.load_workers(self.filename, {"workers": [wrk]})
        self.assertResults(workers=[wrk])

    def test_load_change_sources_defaults(self):
        self.cfg.load_change_sources(self.filename, {})
        self.assertResults(change_sources=[])

    def test_load_change_sources_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_change_sources(self.filename, {'change_source': [mock.Mock()]})

        self.assertConfigError(errors, "must be a list of")

    def test_load_change_sources_single(self):
        chsrc = FakeChangeSource()
        self.cfg.load_change_sources(self.filename, {"change_source": chsrc})
        self.assertResults(change_sources=[chsrc])

    def test_load_change_sources_list(self):
        chsrc = FakeChangeSource()
        self.cfg.load_change_sources(self.filename, {"change_source": [chsrc]})
        self.assertResults(change_sources=[chsrc])

    def test_load_machines_defaults(self):
        self.cfg.load_machines(self.filename, {})
        self.assertResults(machines=[])

    def test_load_machines_not_instance(self):
        with capture_config_errors() as errors:
            self.cfg.load_machines(self.filename, {'machines': [mock.Mock()]})

        self.assertConfigError(errors, "must be a list of")

    def test_load_machines_single(self):
        with capture_config_errors() as errors:
            mm = FakeMachine(name='a')
            self.cfg.load_machines(self.filename, {'machines': mm})

        self.assertConfigError(errors, "must be a list of")

    def test_load_machines_list(self):
        mm = FakeMachine()
        self.cfg.load_machines(self.filename, {"machines": [mm]})
        self.assertResults(machines=[mm])

    def test_load_user_managers_defaults(self):
        self.cfg.load_user_managers(self.filename, {})
        self.assertResults(user_managers=[])

    def test_load_user_managers_not_list(self):
        with capture_config_errors() as errors:
            self.cfg.load_user_managers(self.filename, {'user_managers': 'foo'})

        self.assertConfigError(errors, "must be a list")

    def test_load_user_managers(self):
        um = mock.Mock()
        self.cfg.load_user_managers(self.filename, {"user_managers": [um]})
        self.assertResults(user_managers=[um])

    def test_load_www_default(self):
        self.cfg.load_www(self.filename, {})
        self.assertResults(
            www={
                "port": None,
                "plugins": {},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "logfileName": 'http.log',
            }
        )

    def test_load_www_port(self):
        self.cfg.load_www(self.filename, {"www": {"port": 9888}})
        self.assertResults(
            www={
                "port": 9888,
                "plugins": {},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "logfileName": 'http.log',
            }
        )

    def test_load_www_plugin(self):
        self.cfg.load_www(self.filename, {"www": {"plugins": {'waterfall': {'foo': 'bar'}}}})
        self.assertResults(
            www={
                "port": None,
                "plugins": {'waterfall': {'foo': 'bar'}},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "logfileName": 'http.log',
            }
        )

    def test_load_www_allowed_origins(self):
        self.cfg.load_www(self.filename, {"www": {"allowed_origins": ['a', 'b']}})
        self.assertResults(
            www={
                "port": None,
                "allowed_origins": ['a', 'b'],
                "plugins": {},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "logfileName": 'http.log',
            }
        )

    def test_load_www_logfileName(self):
        self.cfg.load_www(self.filename, {"www": {"logfileName": 'http-access.log'}})
        self.assertResults(
            www={
                "port": None,
                "plugins": {},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "logfileName": 'http-access.log',
            }
        )

    def test_load_www_versions(self):
        custom_versions = [
            ('Test Custom Component', '0.0.1'),
            ('Test Custom Component 2', '0.1.0'),
        ]
        self.cfg.load_www(self.filename, {'www': {"versions": custom_versions}})
        self.assertResults(
            www={
                "port": None,
                "plugins": {},
                "auth": {'name': 'NoAuth'},
                "authz": {},
                "avatar_methods": {'name': 'gravatar'},
                "versions": custom_versions,
                "logfileName": 'http.log',
            }
        )

    def test_load_www_versions_not_list(self):
        with capture_config_errors() as errors:
            custom_versions = {
                'Test Custom Component': '0.0.1',
                'Test Custom Component 2': '0.0.2',
            }
            self.cfg.load_www(self.filename, {'www': {'versions': custom_versions}})
        self.assertConfigError(errors, 'Invalid www configuration value of versions')

    def test_load_www_versions_value_invalid(self):
        with capture_config_errors() as errors:
            custom_versions = [('a', '1'), 'abc', ('b',)]
            self.cfg.load_www(self.filename, {'www': {'versions': custom_versions}})

        self.assertConfigError(errors, 'Invalid www configuration value of versions')

    def test_load_www_cookie_expiration_time_not_timedelta(self):
        with capture_config_errors() as errors:
            self.cfg.load_www(self.filename, {'www': {"cookie_expiration_time": 1}})

        self.assertConfigError(errors, 'Invalid www["cookie_expiration_time"]')

    def test_load_www_unknown(self):
        with capture_config_errors() as errors:
            self.cfg.load_www(self.filename, {"www": {"foo": "bar"}})

        self.assertConfigError(errors, "unknown www configuration parameter(s) foo")

    def test_load_services_nominal(self):
        testcase = self

        class MyService(service.BuildbotService):
            def reconfigService(self, foo=None):
                testcase.foo = foo

        myService = MyService(foo="bar", name="foo")

        self.cfg.load_services(self.filename, {"services": [myService]})
        self.assertResults(services={"foo": myService})

    def test_load_services_badservice(self):
        class MyService:
            pass

        with capture_config_errors() as errors:
            myService = MyService()
            self.cfg.load_services(self.filename, {'services': [myService]})

        errMsg = (
            "<class 'buildbot.test.unit.config.test_master."
            "MasterConfig_loaders.test_load_services_badservice."
            "<locals>.MyService'> "
        )
        errMsg += "object should be an instance of buildbot.util.service.BuildbotService"
        self.assertConfigError(errors, errMsg)

    def test_load_services_duplicate(self):
        with capture_config_errors() as errors:

            class MyService(service.BuildbotService):
                name = 'myservice'

                def reconfigService(self, x=None):
                    self.x = x

            self.cfg.load_services(
                self.filename, {"services": [MyService(x='a'), MyService(x='b')]}
            )

        self.assertConfigError(errors, f'Duplicate service name {MyService.name!r}')

    def test_load_configurators_norminal(self):
        class MyConfigurator(configurators.ConfiguratorBase):
            def configure(self, config_dict):
                config_dict['foo'] = 'bar'

        c = {"configurators": [MyConfigurator()]}
        self.cfg.run_configurators(self.filename, c)
        self.assertEqual(c['foo'], 'bar')


class MasterConfig_checkers(ConfigErrorsMixin, unittest.TestCase):
    def setUp(self):
        self.cfg = config.master.MasterConfig()

    # utils

    def setup_basic_attrs(self):
        # set up a basic config for checking; this will be modified below
        sch = mock.Mock()
        sch.name = 'sch'
        sch.listBuilderNames = lambda: ['b1', 'b2']

        b1 = mock.Mock()
        b1.name = 'b1'

        b2 = mock.Mock()
        b2.name = 'b2'

        self.cfg.schedulers = {"sch": sch}
        self.cfg.workers = [mock.Mock()]
        self.cfg.builders = [b1, b2]

    def setup_builder_locks(
        self, builder_lock=None, dup_builder_lock=False, bare_builder_lock=False
    ):
        """Set-up two mocked builders with specified locks.

        @type  builder_lock: string or None
        @param builder_lock: Name of the lock to add to first builder.
                             If None, no lock is added.

        @type dup_builder_lock: boolean
        @param dup_builder_lock: if True, add a lock with duplicate name
                                 to the second builder

        @type dup_builder_lock: boolean
        @param bare_builder_lock: if True, add bare lock objects, don't wrap
                                  them into locks.LockAccess object
        """

        def bldr(name):
            b = mock.Mock()
            b.name = name
            b.locks = []
            b.factory.steps = [('cls', (), {"locks": []})]
            return b

        def lock(name):
            lock = locks.MasterLock(name)
            if bare_builder_lock:
                return lock
            return locks.LockAccess(lock, "counting", count=1)

        b1 = bldr('b1')
        b2 = bldr('b2')
        self.cfg.builders = [b1, b2]
        if builder_lock:
            b1.locks.append(lock(builder_lock))
            if dup_builder_lock:
                b2.locks.append(lock(builder_lock))

    # tests

    def test_check_single_master_multimaster(self):
        with capture_config_errors() as errors:
            self.cfg.multiMaster = True
            self.cfg.check_single_master()

        self.assertNoConfigErrors(errors)

    def test_check_single_master_no_builders(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            self.cfg.builders = []
            self.cfg.check_single_master()

        self.assertConfigError(errors, "no builders are configured")

    def test_check_single_master_no_workers(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            self.cfg.workers = []
            self.cfg.check_single_master()

        self.assertConfigError(errors, "no workers are configured")

    def test_check_single_master_unsch_builder(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            b3 = mock.Mock()
            b3.name = 'b3'
            self.cfg.builders.append(b3)
            self.cfg.check_single_master()

        self.assertConfigError(errors, "have no schedulers to drive them")

    def test_check_single_master_renderable_builderNames(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            b3 = mock.Mock()
            b3.name = 'b3'
            self.cfg.builders.append(b3)
            sch2 = mock.Mock()
            sch2.listBuilderNames = lambda: properties.Interpolate('%(prop:foo)s')
            self.cfg.schedulers['sch2'] = sch2
            self.cfg.check_single_master()

        self.assertNoConfigErrors(errors)

    def test_check_schedulers_unknown_builder(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            del self.cfg.builders[1]  # remove b2, leaving b1

            self.cfg.check_schedulers()

        self.assertConfigError(errors, "Unknown builder 'b2'")

    def test_check_schedulers_ignored_in_multiMaster(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            del self.cfg.builders[1]  # remove b2, leaving b1
            self.cfg.multiMaster = True
            self.cfg.check_schedulers()

        self.assertNoConfigErrors(errors)

    def test_check_schedulers_renderable_builderNames(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            sch2 = mock.Mock()
            sch2.listBuilderNames = lambda: properties.Interpolate('%(prop:foo)s')
            self.cfg.schedulers['sch2'] = sch2

            self.cfg.check_schedulers()

        self.assertNoConfigErrors(errors)

    def test_check_schedulers(self):
        with capture_config_errors() as errors:
            self.setup_basic_attrs()
            self.cfg.check_schedulers()

        self.assertNoConfigErrors(errors)

    def test_check_locks_dup_builder_lock(self):
        with capture_config_errors() as errors:
            self.setup_builder_locks(builder_lock='l', dup_builder_lock=True)
            self.cfg.check_locks()

        self.assertConfigError(errors, "Two locks share")

    def test_check_locks(self):
        with capture_config_errors() as errors:
            self.setup_builder_locks(builder_lock='bl')
            self.cfg.check_locks()

        self.assertNoConfigErrors(errors)

    def test_check_locks_none(self):
        # no locks in the whole config, should be fine
        with capture_config_errors() as errors:
            self.setup_builder_locks()
            self.cfg.check_locks()

        self.assertNoConfigErrors(errors)

    def test_check_locks_bare(self):
        # check_locks() should be able to handle bare lock object,
        # lock objects that are not wrapped into LockAccess() object

        with capture_config_errors() as errors:
            self.setup_builder_locks(builder_lock='oldlock', bare_builder_lock=True)
            self.cfg.check_locks()

        self.assertNoConfigErrors(errors)

    def test_check_builders_unknown_worker(self):
        with capture_config_errors() as errors:
            wrk = mock.Mock()
            wrk.workername = 'xyz'
            self.cfg.workers = [wrk]

            b1 = FakeBuilder(workernames=['xyz', 'abc'], builddir='x', name='b1', project=None)
            self.cfg.builders = [b1]

            self.cfg.check_builders()
        self.assertConfigError(errors, "builder 'b1' uses unknown workers 'abc'")

    def test_check_builders_unknown_project(self):
        with capture_config_errors() as errors:
            wrk = mock.Mock()
            wrk.name = 'proj1'
            self.cfg.projects = [wrk]

            b1 = FakeBuilder(workernames=[], builddir='x', name='b1', project='proj_unknown')
            self.cfg.builders = [b1]

            self.cfg.check_builders()
        self.assertConfigError(errors, "builder 'b1' uses unknown project name 'proj_unknown'")

    def test_check_builders_duplicate_name(self):
        with capture_config_errors() as errors:
            b1 = FakeBuilder(workernames=[], name='b1', builddir='1', project=None)
            b2 = FakeBuilder(workernames=[], name='b1', builddir='2', project=None)
            self.cfg.builders = [b1, b2]

            self.cfg.check_builders()

        self.assertConfigError(errors, "duplicate builder name 'b1'")

    def test_check_builders_duplicate_builddir(self):
        with capture_config_errors() as errors:
            b1 = FakeBuilder(workernames=[], name='b1', builddir='dir', project=None)
            b2 = FakeBuilder(workernames=[], name='b2', builddir='dir', project=None)
            self.cfg.builders = [b1, b2]

            self.cfg.check_builders()

        self.assertConfigError(errors, "duplicate builder builddir 'dir'")

    def test_check_builders(self):
        with capture_config_errors() as errors:
            wrk = mock.Mock()
            wrk.workername = 'a'
            self.cfg.workers = [wrk]

            project = mock.Mock()
            project.name = 'proj1'
            self.cfg.projects = [project]

            b1 = FakeBuilder(workernames=['a'], name='b1', builddir='dir1', project='proj1')
            b2 = FakeBuilder(workernames=['a'], name='b2', builddir='dir2', project=None)
            self.cfg.builders = [b1, b2]

            self.cfg.check_builders()

        self.assertNoConfigErrors(errors)

    def test_check_duplicate_projects(self):
        with capture_config_errors() as errors:
            project1 = mock.Mock()
            project1.name = 'proj1'
            project2 = mock.Mock()
            project2.name = 'proj1'
            self.cfg.projects = [project1, project2]

            self.cfg.check_projects()

        self.assertConfigError(errors, "duplicate project name 'proj1'")

    def test_check_ports_protocols_set(self):
        with capture_config_errors() as errors:
            self.cfg.protocols = {"pb": {"port": 10}}
            self.cfg.check_ports()

        self.assertNoConfigErrors(errors)

    def test_check_ports_protocols_not_set_workers(self):
        with capture_config_errors() as errors:
            self.cfg.workers = [mock.Mock()]
            self.cfg.check_ports()

        self.assertConfigError(errors, "workers are configured, but c['protocols'] not")

    def test_check_ports_protocols_port_duplication(self):
        with capture_config_errors() as errors:
            self.cfg.protocols = {"pb": {"port": 123}, "amp": {"port": 123}}
            self.cfg.check_ports()

        self.assertConfigError(errors, "Some of ports in c['protocols'] duplicated")

    def test_check_machines_unknown_name(self):
        with capture_config_errors() as errors:
            self.cfg.workers = [FakeWorker(name='wa', machine_name='unk')]
            self.cfg.machines = [FakeMachine(name='a')]
            self.cfg.check_machines()

        self.assertConfigError(errors, 'uses unknown machine')

    def test_check_machines_duplicate_name(self):
        with capture_config_errors() as errors:
            self.cfg.machines = [FakeMachine(name='a'), FakeMachine(name='a')]
            self.cfg.check_machines()

        self.assertConfigError(errors, 'duplicate machine name')


class MasterConfig_old_worker_api(unittest.TestCase):
    filename = "test.cfg"

    def setUp(self):
        self.cfg = config.master.MasterConfig()

    def test_workers_new_api(self):
        with assertNotProducesWarnings(DeprecatedApiWarning):
            self.assertEqual(self.cfg.workers, [])


class FakeService(service.ReconfigurableServiceMixin, service.AsyncService):
    succeed = True
    call_index = 1

    @defer.inlineCallbacks
    def reconfigServiceWithBuildbotConfig(self, new_config):
        self.called = FakeService.call_index
        FakeService.call_index += 1
        yield super().reconfigServiceWithBuildbotConfig(new_config)
        if not self.succeed:
            raise ValueError("oh noes")


class FakeMultiService(service.ReconfigurableServiceMixin, service.AsyncMultiService):
    def reconfigServiceWithBuildbotConfig(self, new_config):
        self.called = True
        d = super().reconfigServiceWithBuildbotConfig(new_config)
        return d


class ReconfigurableServiceMixin(unittest.TestCase):
    @defer.inlineCallbacks
    def test_service(self):
        svc = FakeService()
        yield svc.reconfigServiceWithBuildbotConfig(mock.Mock())

        self.assertTrue(svc.called)

    @defer.inlineCallbacks
    def test_service_failure(self):
        svc = FakeService()
        svc.succeed = False
        try:
            yield svc.reconfigServiceWithBuildbotConfig(mock.Mock())
        except ValueError:
            pass
        else:
            self.fail("should have raised ValueError")

    @defer.inlineCallbacks
    def test_multiservice(self):
        svc = FakeMultiService()
        ch1 = FakeService()
        yield ch1.setServiceParent(svc)
        ch2 = FakeMultiService()
        yield ch2.setServiceParent(svc)
        ch3 = FakeService()
        yield ch3.setServiceParent(ch2)
        yield svc.reconfigServiceWithBuildbotConfig(mock.Mock())

        self.assertTrue(svc.called)
        self.assertTrue(ch1.called)
        self.assertTrue(ch2.called)
        self.assertTrue(ch3.called)

    @defer.inlineCallbacks
    def test_multiservice_priority(self):
        parent = FakeMultiService()
        svc128 = FakeService()
        yield svc128.setServiceParent(parent)

        services = [svc128]
        for i in range(20, 1, -1):
            svc = FakeService()
            svc.reconfig_priority = i
            yield svc.setServiceParent(parent)
            services.append(svc)

        yield parent.reconfigServiceWithBuildbotConfig(mock.Mock())

        prio_order = [s.called for s in services]
        called_order = sorted(prio_order)
        self.assertEqual(prio_order, called_order)

    @defer.inlineCallbacks
    def test_multiservice_nested_failure(self):
        svc = FakeMultiService()
        ch1 = FakeService()
        yield ch1.setServiceParent(svc)
        ch1.succeed = False
        try:
            yield svc.reconfigServiceWithBuildbotConfig(mock.Mock())
        except ValueError:
            pass
        else:
            self.fail("should have raised ValueError")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import test_results
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class Tests(TestReactorMixin, unittest.TestCase):
    common_data = [
        fakedb.Worker(id=47, name='linux'),
        fakedb.Buildset(id=20),
        fakedb.Builder(id=88, name='b1'),
        fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
        fakedb.Master(id=88),
        fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47),
        fakedb.Step(id=131, number=132, name='step132', buildid=30),
        fakedb.TestResultSet(
            id=13,
            builderid=88,
            buildid=30,
            stepid=131,
            description='desc',
            category='cat',
            value_unit='ms',
            complete=1,
        ),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_add_set_results(self):
        yield self.db.insert_test_data(self.common_data)

        result_values = [
            {'test_name': 'name1', 'value': '1'},
            {'test_name': 'name1', 'duration_ns': 1000, 'value': '2'},
            {'test_name': 'name2', 'test_code_path': 'path2', 'value': '3'},
            {'test_name': 'name3', 'test_code_path': 'path3', 'value': '4'},
            {'test_name': 'name4', 'test_code_path': 'path4', 'line': 4, 'value': '5'},
            {'test_code_path': 'path5', 'line': 5, 'value': '6'},
        ]

        yield self.db.test_results.addTestResults(
            builderid=88, test_result_setid=13, result_values=result_values
        )

        result_dicts = yield self.db.test_results.getTestResults(builderid=88, test_result_setid=13)
        for d in result_dicts:
            self.assertIsInstance(d, test_results.TestResultModel)

        result_dicts = sorted(result_dicts, key=lambda x: x.id)
        resultid = result_dicts[0].id
        self.assertEqual(
            result_dicts,
            [
                test_results.TestResultModel(
                    id=resultid,
                    builderid=88,
                    test_result_setid=13,
                    test_name='name1',
                    test_code_path=None,
                    line=None,
                    duration_ns=None,
                    value='1',
                ),
                test_results.TestResultModel(
                    id=resultid + 1,
                    builderid=88,
                    test_result_setid=13,
                    test_name='name1',
                    test_code_path=None,
                    line=None,
                    duration_ns=1000,
                    value='2',
                ),
                test_results.TestResultModel(
                    id=resultid + 2,
                    builderid=88,
                    test_result_setid=13,
                    test_name='name2',
                    test_code_path='path2',
                    line=None,
                    duration_ns=None,
                    value='3',
                ),
                test_results.TestResultModel(
                    id=resultid + 3,
                    builderid=88,
                    test_result_setid=13,
                    test_name='name3',
                    test_code_path='path3',
                    line=None,
                    duration_ns=None,
                    value='4',
                ),
                test_results.TestResultModel(
                    id=resultid + 4,
                    builderid=88,
                    test_result_setid=13,
                    test_name='name4',
                    test_code_path='path4',
                    line=4,
                    duration_ns=None,
                    value='5',
                ),
                test_results.TestResultModel(
                    id=resultid + 5,
                    builderid=88,
                    test_result_setid=13,
                    test_name=None,
                    test_code_path='path5',
                    line=5,
                    duration_ns=None,
                    value='6',
                ),
            ],
        )

        result_dict = yield self.db.test_results.getTestResult(test_resultid=resultid)
        self.assertEqual(
            result_dict,
            test_results.TestResultModel(
                id=resultid,
                builderid=88,
                test_result_setid=13,
                test_name='name1',
                test_code_path=None,
                line=None,
                duration_ns=None,
                value='1',
            ),
        )

    @defer.inlineCallbacks
    def test_get_names(self):
        yield self.db.insert_test_data([
            *self.common_data,
            fakedb.TestName(id=103, builderid=88, name='name103'),
            fakedb.TestName(id=104, builderid=88, name='name104'),
            fakedb.TestName(id=105, builderid=88, name='name105'),
            fakedb.TestName(id=116, builderid=88, name='name116'),
            fakedb.TestName(id=117, builderid=88, name='name117'),
        ])

        name_dicts = yield self.db.test_results.getTestNames(builderid=88)
        self.assertEqual(name_dicts, ['name103', 'name104', 'name105', 'name116', 'name117'])

        name_dicts = yield self.db.test_results.getTestNames(
            builderid=88, name_prefix='non_existing'
        )
        self.assertEqual(name_dicts, [])

        name_dicts = yield self.db.test_results.getTestNames(builderid=88, name_prefix='name10')
        self.assertEqual(name_dicts, ['name103', 'name104', 'name105'])

        name_dicts = yield self.db.test_results.getTestNames(builderid=88, name_prefix='name11')
        self.assertEqual(name_dicts, ['name116', 'name117'])

    @defer.inlineCallbacks
    def test_get_code_paths(self):
        yield self.db.insert_test_data([
            *self.common_data,
            fakedb.TestCodePath(id=103, builderid=88, path='path103'),
            fakedb.TestCodePath(id=104, builderid=88, path='path104'),
            fakedb.TestCodePath(id=105, builderid=88, path='path105'),
            fakedb.TestCodePath(id=116, builderid=88, path='path116'),
            fakedb.TestCodePath(id=117, builderid=88, path='path117'),
        ])

        path_dicts = yield self.db.test_results.getTestCodePaths(builderid=88)
        self.assertEqual(path_dicts, ['path103', 'path104', 'path105', 'path116', 'path117'])

        path_dicts = yield self.db.test_results.getTestCodePaths(
            builderid=88, path_prefix='non_existing'
        )
        self.assertEqual(path_dicts, [])

        path_dicts = yield self.db.test_results.getTestCodePaths(builderid=88, path_prefix='path10')
        self.assertEqual(path_dicts, ['path103', 'path104', 'path105'])

        path_dicts = yield self.db.test_results.getTestCodePaths(builderid=88, path_prefix='path11')
        self.assertEqual(path_dicts, ['path116', 'path117'])
